from typing import List
import pandas as pd

from sklearn.preprocessing import MultiLabelBinarizer


def validate_multiple_labels(y_raw: List[List]) -> None:
    """
    Validate binarization of labels in a multi-label setting.

    Args:
        y_raw: Raw list of list of labels.

    Returns:

    """
    y = MultiLabelBinarizer().fit_transform(y_raw)

    # Assert that every sample has atleast one label
    assert (y.sum(axis=1) == 0).sum() == 0

    # Assert that every label is assigned to some data point
    assert (y.sum(axis=0) == 0).sum() == 0

    # Assert that no label is assigned to only one data point
    assert not (y.sum(axis=0) == 1).any()


def count_single_label(y_raw: List[List]) -> pd.DataFrame:
    binarizer = MultiLabelBinarizer()
    y = binarizer.fit_transform(y_raw)
    label_counts = y.sum(axis=0)
    label_names = binarizer.classes_
    label_df = pd.DataFrame({'label': label_names,
                             'count': label_counts})
    return label_df.sort_values(by='count', ascending=False)



import base64
import csv


def create_download_link(dataframe,
                         filename: str,
                         file_type: str = 'csv',
                         index: bool = False,
                         header: bool = True):
    """
    Generate a download link for a pandas dataframe.

    Args:
        dataframe: Pandas DataFrame
        filename: Name of exported file
        file_type: Either 'csv' or 'tsv'
        index: Whether to include index of dataframe or not
        header: Whether to include header of dataframe or not

    Returns:
        Markdown to place in st.markdown(...)
    """
    if file_type == 'csv':
        dataframe_csv = dataframe.to_csv(index=index)
    elif file_type == 'tsv':
        dataframe_csv = dataframe.to_csv(index=index,
                                         sep='\t',
                                         header=header,
                                         quoting=csv.QUOTE_NONNUMERIC)
    else:
        raise Exception('Invalid file_type. Allowed values are "csv" and "tsv".')

    b64 = base64.b64encode(dataframe_csv.encode()).decode()
    href = f'**DOWNLOAD:** <a href="data:file/csv;base64,{b64}" download="{filename}">{filename}</a>'
    return href



class ValidationError(Exception):
    """
    A class for validation errors.

    """

    def __init__(self, message):
        super().__init__(message)
        self.message = message

    def __unicode__(self):
        return self.message

    def __str__(self):
        return self.message.encode("utf-8")



from pathlib import Path
from typing import List, Union
import time
import pandas as pd


def print_markdown(markdown):
    from IPython.display import Markdown, display
    display(Markdown(markdown))


def print_bullets(lines: List[str]) -> None:
    """
    Display a list of text as bullet points.

    Args:
        lines: List of texts

    Returns:
        None
    """
    bullet_points = '\n'.join(f'- `{line}`' for line in sorted(lines))
    print_markdown(bullet_points)


def print_header(text: str,
                 level: int = 2) -> None:
    """
    Display a text as markdown header.

    Args:
        text: Text
        level: 2 for H2, 3 for H3 upto 6.

    Returns:
        None
    """
    print_markdown(f'{"#" * level} {text}')


def highlight_phrases(original_text: str,
                      phrases: Union[List[str], str],
                      color_palette: str = 'Greens',
                      weight: float = 0.2) -> None:
    """
    Highlight a list of phrases in a text.

    Args:
        original_text: Sentence
        phrases: A single phrase or a list of phrases
        color_palette: Any valid matplotlib color palette name
        weight: Darkness of the color

    Returns:
        None
    """
    import matplotlib.cm
    from IPython.display import HTML, display

    html = original_text
    cmap = matplotlib.cm.get_cmap(color_palette)
    color = f'rgba{cmap(weight, bytes=True)}'
    if type(phrases) is str:
        phrases = [phrases]
    for phrase in phrases:
        highlighted_phrase = (f'<span style="background-color: {color}; font-weight: {weight * 800};">'
                              f'{phrase}'
                              f'</span>')
        html = html.replace(phrase, highlighted_phrase)
    display(HTML(f'<p style="color: #444; font-size:1.5em;">{html}</p>'))


def filter_column(df: pd.DataFrame,
                  column_name: str) -> None:
    """
    Show an interactive widget to filter a column in dataframe.

    Args:
        df: Pandas DataFrame
        column_name: Column Name of the DataFrame

    Returns:
        Interactive widget for filtering.
    """

    from ipywidgets import interact
    options = sorted(df[column_name].unique())
    interact(lambda value: df[df[column_name] == value],
             value=options)


def download(file_path) -> None:
    """
    Download a file at given path.

    Args:
        file_path: File path

    Returns:
        None
    """
    from IPython.display import Javascript
    script = f'''
            var host = window.location.host;
            var downloadLink = window.location.protocol + "//" + host + "/files/{file_path}"
            window.open(downloadLink)
            '''
    return Javascript(script)


def download_df(df: pd.DataFrame,
                csv_path=None) -> None:
    """
    Download a dataframe as a CSV with a random filename.

    The filename is set to a random UUID.

    Args:
        df: Pandas DataFrame
        csv_path: CSV filename.

    Returns:
        None
    """
    from IPython.display import display
    if not csv_path:
        from uuid import uuid4
        csv_path = f'{uuid4()}.csv'
    df.to_csv(csv_path, index=False)
    display(download(file_path=csv_path))
    time.sleep(1)
    Path(csv_path).unlink()


def search_dataframe(df: pd.DataFrame) -> None:
    """
    Show an interactive widget to search text fields of a dataframe.

    Args:
        df: Pandas DataFrame

    Returns:
        Interactive widget for searching.
    """

    from ipywidgets import interact
    from IPython.display import display

    def _search(query: str, column: str):
        if query:
            with pd.option_context('display.max_rows', None,
                                   'display.max_columns', None):
                filtered_df = df[df[column].str.contains(query,
                                                         case=False,
                                                         regex=False)]
                display(filtered_df)

    string_columns = df.select_dtypes('object').columns.tolist()
    interact(_search, query='', column=string_columns)


def show_examples(df: pd.DataFrame,
                  group_column: str,
                  data_column: str,
                  n: int = 5):
    """
    Show random examples for each sub-group in a dataframe.

    Args:
        df: Dataframe
        group_column: Column name for performing group by
        data_column: Column to show examples for
        n: Number of examples

    Returns:
        Markdown
    """
    from IPython.display import Markdown
    generated_text = ""
    for group_name, subset in df.explode(group_column).groupby(group_column):
        examples = subset[data_column].sample(n)
        generated_text += f"## {group_name}\n\n"
        generated_text += "\n".join([f'- {example}' for example in examples])
        generated_text += '\n\n'
    return Markdown(generated_text)



from fns.cluster import *
from fns.constants import *
from fns.dataframe import *
from fns.decorators import *
from fns.exceptions import *
from fns.json_encoders import *
from fns.metrics import *
from fns.model_selection import *
from fns.preprocessing import *
from fns.streamlit_utils import *
from fns.text import *
from fns.vision import *
from fns.fns import *
from fns.plot import *
from fns.multi_label import *
from sklearn.datasets import (make_classification,
                              make_regression,
                              make_multilabel_classification)
from sklearn.dummy import DummyRegressor, DummyClassifier
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor, ExtraTreeClassifier, ExtraTreeRegressor
from sklearn.linear_model import (LogisticRegression,
                                  LogisticRegressionCV,
                                  LinearRegression,
                                  SGDRegressor,
                                  SGDClassifier,
                                  RidgeClassifier,
                                  RidgeClassifierCV,
                                  Lasso,
                                  LassoCV,
                                  ElasticNet,
                                  Ridge,
                                  RidgeCV,
                                  )
from sklearn.naive_bayes import GaussianNB, BernoulliNB, MultinomialNB
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.neighbors import (KNeighborsClassifier,
                               RadiusNeighborsClassifier,
                               RadiusNeighborsRegressor)
from sklearn.svm import SVC, LinearSVC, LinearSVR, SVR, OneClassSVM
from sklearn.ensemble import (RandomForestClassifier,
                              RandomForestRegressor,
                              ExtraTreesClassifier,
                              ExtraTreesRegressor,
                              BaggingClassifier,
                              BaggingRegressor,
                              AdaBoostClassifier,
                              AdaBoostRegressor,
                              GradientBoostingClassifier,
                              GradientBoostingRegressor)
from sklearn.cluster import KMeans, AffinityPropagation, AgglomerativeClustering
from sklearn.neural_network import MLPClassifier
from sklearn.decomposition import PCA, TruncatedSVD, NMF, LatentDirichletAllocation
from sklearn.manifold import TSNE
from sklearn.model_selection import train_test_split, GridSearchCV, RandomizedSearchCV, cross_val_score
from sklearn.model_selection import StratifiedKFold, KFold, ParameterGrid, cross_val_predict
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer, HashingVectorizer, TfidfTransformer
from sklearn.feature_extraction.text import ENGLISH_STOP_WORDS
from sklearn.preprocessing import (normalize,
                                   scale,
                                   minmax_scale,
                                   RobustScaler,
                                   MinMaxScaler,
                                   MaxAbsScaler,
                                   OneHotEncoder,
                                   FunctionTransformer,
                                   PowerTransformer,
                                   StandardScaler,
                                   MultiLabelBinarizer,
                                   LabelBinarizer,
                                   Normalizer,
                                   PolynomialFeatures,
                                   )
from sklearn.multiclass import (OneVsOneClassifier,
                                OneVsRestClassifier)
from sklearn.metrics import (accuracy_score,
                             precision_score,
                             recall_score,
                             confusion_matrix,
                             silhouette_score,
                             classification_report,
                             f1_score,
                             fbeta_score,
                             mean_squared_error,
                             mean_absolute_error,
                             matthews_corrcoef,
                             hamming_loss)
from sklearn.metrics.pairwise import cosine_distances, cosine_similarity, euclidean_distances
from sklearn.pipeline import Pipeline, FeatureUnion, make_pipeline
from sklearn.calibration import calibration_curve, CalibratedClassifierCV
import matplotlib.pyplot as plt
from itertools import chain, product, permutations, combinations, islice
from functools import lru_cache
from collections import Counter, OrderedDict, defaultdict
from pathlib import Path
import math
import random

imports = """
import seaborn as sns
import torch
import torch.optim as optim
import torch.nn as nn
from torch.utils.data import Dataset, TensorDataset, DataLoader
from torch.utils.tensorboard import SummaryWriter
import nltk
import spacy
import requests
from textblob import TextBlob
import ipywidgets as widgets
from pprint import pprint
from typing import List, Dict, Tuple, Union, Optional
from loguru import logger
from fns.notebook import *
from sklearn.inspection import permutation_importance
from tqdm.auto import tqdm
""".strip().splitlines()

for import_line in imports:
    try:
        exec(import_line)
    except Exception:
        pass



from pathlib import Path
from typing import Dict, List, Union

import pandas as pd
import numpy as np


def fake_df() -> pd.DataFrame:
    """
    Generate a dataframe filled with random data.

    Returns:
        Pandas DataFrame
    """
    return pd.util.testing.makeDataFrame()


def read_dict(data: Dict) -> pd.DataFrame:
    """
    Create a dataframe from dictionary with unequal elements.

    Args:
        data: Dictionary with column names as keys and rows as values

    Returns:
        Pandas DataFrame
    """
    return pd.DataFrame.from_dict(data, orient='index').transpose()


def print_groups(df: pd.DataFrame, column: str) -> None:
    """
    Pretty print all subsets of a groupby.

    Args:
        df: Pandas DataFrame
        column: Column Name to group by

    Returns:
        None
    """
    for current_group, sub_df in df.groupby(column):
        print(f'Group: {current_group}')
        print()
        # Skip group column
        mask = ~(sub_df.columns.isin([column]))
        print(sub_df.loc[:, mask])
        print()
        print('---' * 25)


def display_all() -> None:
    """
    Show all the rows and columns when printing dataframe.

    Returns:
        None
    """
    import pandas as pd
    pd.set_option('display.max_colwidth', None)
    pd.set_option('display.max_rows', None)


def no_wrapping():
    """
    Return a context manager to display all rows and columns.

    Examples:
    ```python
    with no_wrapping():
        print(df)
    ```

    Returns:
        Context Manager
    """
    return pd.option_context('display.max_rows', None,
                             'display.max_columns', None)


def explore_df(df: pd.DataFrame) -> pd.DataFrame:
    """
    Perform a quick peek of a dataframe.

    Currently shows:
    - Number of null elements in each column
    - Data type of each column
    - One example data for each column

    Args:
        df: Pandas DataFrame

    Returns:
        DataFrame with summary infos
    """
    null_df = pd.DataFrame(df.isnull().sum(), columns=['num_nulls'])
    dtype_df = pd.DataFrame(df.dtypes, columns=['dtype'])
    return (df.T.sample(1, axis=1)
            .join([dtype_df, null_df])
            .rename_axis('Columns')
            )


def is_outlier(values: List) -> List[bool]:
    """
    Generate a mask if an element is an outlier or not.

    Extra:
    ```
    Condition 1: < Q1 - 1.5 * IQR
    Condition 2: > Q3 + 1.5 * IQR
    ```

    Args:
        values: List of numerical values

    Returns:
        List of boolean indicating if an element is outlier or not
    """
    q1 = np.quantile(values, 0.25)
    q3 = np.quantile(values, 0.75)
    iqr = q3 - q1
    lower_threshold = q1 - 1.5 * iqr
    upper_threshold = q3 + 1.5 * iqr
    return (values < lower_threshold) | (values > upper_threshold)


def to_excel(path: Union[Path, str],
             df: pd.DataFrame,
             sheet_name: str,
             index: bool = False,
             mode: str = 'a') -> None:
    """
    Add a dataframe to an existing Excel file.

    Args:
        path: Path of the excel file
        df: Pandas DataFrame
        sheet_name: The sheet name to save in
        index: Keep or remove index
        mode: 'a' for append or 'w' for write

    Returns:
        None
    """
    with pd.ExcelWriter(path, mode=mode) as writer:
        df.to_excel(writer,
                    sheet_name=sheet_name,
                    index=index)



import time
from typing import Callable
import functools

from fns import minibatch


def timeit(func: Callable) -> Callable:
    """
    Decorator to calculate time taken for a function to complete.

    Args:
        func: Python Function

    Returns:
        Decorated function
    """
    start_time = time.time()

    @functools.wraps(func)
    def inner(*args, **kwargs):
        func(*args, **kwargs)
        total_time_taken = time.time() - start_time
        print('Total time taken: {} seconds'.format(total_time_taken))

    return inner


def show_shapes(func: Callable) -> Callable:
    """
    Decorator to log dataframe shape before and after applying a function.

    Args:
        func: Function that takes a dataframe as argument

    Returns:
        function
    """

    @functools.wraps(func)
    def inner(df):
        print(f'Shape before {func.__name__}', df.shape)
        out_df = func(df)
        print(f'Shape after {func.__name__}', out_df.shape)
        return out_df

    return inner


def deduplicate(func: Callable) -> Callable:
    """
    Decorator to deduplicate results of a function.

    Usage:
    ```python
    @deduplicate
    def test():
        return [1, 2, 3, 1]
    ```

    Args:
        func: Function

    Returns:
        Function
    """

    @functools.wraps(func)
    def inner(*args, **kwargs):
        return list(set(func(*args, **kwargs)))

    return inner


def batched(batch_size: int = 32) -> Callable:
    """
    Apply a function over small batches of a list and combine results.

    Args:
        batch_size: Size of each mini-batch

    Returns:
        Decorator for the batch size
    """

    def decorator(func) -> Callable:
        @functools.wraps(func)
        def inner(*args, **kwargs):
            items = args[0]
            results = []
            for batch in minibatch(items, batch_size):
                batch_results = func(batch)
                results.extend(batch_results)
            return results

        return inner

    return decorator


def to(data_type) -> Callable:
    """
    Apply a data type to returned data from a function.

    Args:
        data_type: The data type to apply. Eg: list, int etc.

    Returns:
        Decorator that applies the data type on returned data
    """

    def decorator(func) -> Callable:
        @functools.wraps(func)
        def inner(*args, **kwargs):
            return data_type(func(*args, **kwargs))

        return inner

    return decorator



import math
from collections import Counter
from typing import Dict, Callable, List, Tuple

import numpy as np
import timeit

import pandas as pd
from sklearn import metrics as M


def clustering_report(y_true, y_pred) -> pd.DataFrame:
    """
    Generate cluster evaluation metrics.


    Args:
        y_true: Array of actual labels
        y_pred: Array of predicted clusters

    Returns:
        Pandas DataFrame with metrics.
    """
    return pd.DataFrame(
        {
            "Homogeneity": M.homogeneity_score(y_true, y_pred),
            "Completeness": M.completeness_score(y_true, y_pred),
            "V-Measure": M.v_measure_score(y_true, y_pred),
            "Adjusted Rand Index": M.adjusted_rand_score(y_true, y_pred),
            "Adjusted Mutual Information": M.adjusted_mutual_info_score(y_true, y_pred),
        },
        index=["value"],
    ).T


def multilabel_classification_report(y_true, y_pred) -> pd.Series:
    """
    Compute all metrics for a multi-label classification problem.
    Args:
        y_true: True binarized labels
        y_pred: Predicted binarized labels

    Returns:
        Pandas series of metrics
    """
    scores = {
        "accuracy": M.accuracy_score(y_true, y_pred),
        "precision_macro": M.precision_score(y_true, y_pred, average="macro"),
        "recall_macro": M.recall_score(y_true, y_pred, average="macro"),
        "f1_samples": M.f1_score(y_true, y_pred, average="samples"),
        "f1_macro": M.f1_score(y_true, y_pred, average="macro"),
        "f1_weighted": M.f1_score(y_true, y_pred, average="weighted"),
        "hamming_loss": M.hamming_loss(y_true, y_pred),
    }
    return pd.Series(scores)


def benchmark_function(fn: Callable, repeat: int = 5) -> Dict:
    """
    Benchmark time taken for a function and return metrics.

    Args:
        fn: A python function
        repeat: Number of samples

    Returns:
        Dictionary of total times, mean and std of times
    """
    iteration_times = timeit.repeat(fn, repeat=repeat, number=1)
    return {
        "time": iteration_times,
        "mean": np.mean(iteration_times),
        "std": np.std(iteration_times),
    }


def baseline_accuracy(labels: List) -> float:
    """
    Get accuracy for always majority class classifier.

    Usage:
    ```python
    >>> baseline_accuracy([0, 1])
    50.0
    ```

    Args:
        labels: List of class labels.

    Returns:
        Baseline accuracy
    """
    (label, count), *_ = Counter(labels).most_common(1)
    return count / len(labels) * 100.0


def missing_value_percent(df: pd.DataFrame) -> pd.DataFrame:
    """
    Get the percentage of missing values in each column.

    Args:
        df: Pandas DataFrame

    Returns:
        Percentage of missing value in each column.
    """
    num_rows = len(df)
    return (df.isna().sum() / num_rows * 100.0).sort_values(ascending=False)


def na_percent(df: pd.DataFrame) -> pd.DataFrame:
    return missing_value_percent(df)


def n_clusters(data) -> int:
    """
    Generate number of clusters to create.

    Heuristic:
    Number of clusters = square root of total data points

    Args:
        data: Total number of data points or the data point itself

    Returns:
        Number of clusters
    """
    if type(data) is int:
        total_rows = data
    else:
        total_rows = len(set(data))
    return int(math.sqrt(total_rows))


def vector_similarity(u, v) -> float:
    angle = np.dot(u, v) / (np.linalg.norm(u) * np.linalg.norm(v))
    return float(angle)


def jaccard(x, y) -> float:
    """
    Compute jaccard similarity (intersection over union).

    Args:
        x: Array-like object
        y: Array-like object

    Returns:
        Intersection Over Union score
    """
    s1 = set(x)
    s2 = set(y)
    if len(s1) == 0 and len(s2) == 0:
        return 0
    return len(s1 & s2) / len(s1 | s2)


def outlier_cutoff(values: List) -> Tuple[float, float]:
    """
    Generate the lower and upper bound for outliers.

    Extra:
    ```
    Lower bound: < Q1 - 1.5 * IQR
    Upper bound: > Q3 + 1.5 * IQR
    ```

    Args:
        values: List of numerical values

    Returns:
        Tuple of (lower-cutoff, upper-cutoff)
    """
    q1 = np.quantile(values, 0.25)
    q3 = np.quantile(values, 0.75)
    iqr = q3 - q1
    lower_threshold = q1 - 1.5 * iqr
    upper_threshold = q3 + 1.5 * iqr
    return lower_threshold, upper_threshold


def sorted_classification_report(y_true, y_pred, **kwargs) -> pd.DataFrame:
    """
    Generate class-wise classification report sorted from worst to best.

    Args:
        y_true: Actual labels
        y_pred: Predicted labels

    Returns:
        Classification report in sorted form.
    """
    base_report = M.classification_report(y_true, y_pred, output_dict=True, **kwargs)
    base_report_df = pd.DataFrame.from_dict(base_report).T
    class_wise_df = base_report_df.iloc[:-3].sort_values(by="f1-score")
    summary_df = base_report_df.iloc[-3:]
    combined_df = pd.concat([class_wise_df, summary_df])
    combined_df["support"] = combined_df["support"].astype(int)
    return combined_df



import argparse
import base64
import itertools
import json
import math
import os
import pickle
import time
from collections import Counter
from pathlib import Path
from typing import List, Dict, Union, Iterator, Any, IO

from fns.text import md5_hash


def flatten(x: List[List]) -> Iterator:
    """
    Flatten a list of list.

    Args:
        x: List of list of elements

    Returns:
        Iterator of flattened array.
    """
    return itertools.chain.from_iterable(x)


def array_except_element(arr: List, elem: Any) -> List:
    """
    Get copy of array without an element.

    Args:
        arr:
        elem:

    Returns:
        Array

    Example:
    ```python
    >>> array_except_element([1, 2, 3], 3)
    [1, 2]
    ```
    """
    elem_index = arr.index(elem)
    return arr[:elem_index] + arr[elem_index + 1:]


def sort_dict_by_value(d: Dict,
                       reverse: bool = False) -> Dict:
    """
    Sort items in dictionary by value.

    Example:
    ```python
    >>> sort_dict_by_value({'gold': 40, 'silver': 25})
    {'silver': 25, 'gold': 40}
    ```

    Args:
        d: Python Dictionary
        reverse: Sort order

    Returns:
        Sorted dictionary
    """
    return dict(sorted(d.items(), key=lambda item: item[1], reverse=reverse))


def reverse_mapping(d: Dict) -> Dict:
    """
    Swap mapping from key: value to value: key

    Args:
        d: Python Dictionary

    Returns:
        Dictionary with key and value swapped
    """
    return {v: k for k, v in d.items()}


def percent_dict(d: Dict) -> Dict:
    """
    Convert a dictionary of key-value to key:coverage-percent.

    Args:
        d: Dictionary of key and values

    Returns:
        Dictionary of key and percent-coverage
    """
    total = sum(d.values())
    return {key: value / total * 100.0
            for key, value in d.items()}


def top(data, n: int = 5) -> Dict:
    """
    Get a dictionary of top-n items from a list.

    Args:
        data: Python collection
        n: Number of top-values

    Returns:
        Dictionary of top-n items and count
    """
    return dict(Counter(data).most_common(n))


def top_n_from_dict(dictionary: Dict,
                    n: int = 10):
    """
    Get top n largest values from the dictionary.

    Args:
        dictionary: Python dictionary
        n: Number of keys to pick

    Returns:

    """
    return top(dictionary, n=n)


def read_json(json_path: Union[str, Path]) -> Dict:
    """
    Read json file from a path.

    Args:
        json_path: File path to a json file.

    Returns:
        Python dictionary
    """
    with open(json_path, 'r') as fp:
        data = json.load(fp)
    return data


def write_json(item: Dict,
               path: Union[Path, str],
               mode: str = 'w') -> None:
    """
    Save json to a file.

    Args:
        item: Python dictionary
        path: File path to save at
        mode: File write mode

    Returns:
        None
    """
    with open(path, mode=mode) as fp:
        json.dump(item, fp)


def read_pickle(path: Union[str, Path]) -> Any:
    """
    Read a pickle file from path.

    Args:
        path: File path

    Returns:
        Unpickled object
    """
    with open(path, 'rb') as fp:
        return pickle.load(fp)


def write_pickle(item: Any,
                 path: Union[Path, str]) -> None:
    """
    Pickle a python object.

    Args:
        item: Python object
        path: File path to save the pickle file

    Returns:
        None
    """
    with open(path, 'wb') as fp:
        pickle.dump(item, fp)


def parse_manual(parser: argparse.ArgumentParser,
                 command: str) -> argparse.Namespace:
    """
    Use argument parser in notebooks.

    Args:
        parser: ArgumentParser
        command: Command line arguments as string

    Returns:
        Parsed argument as namespace
    """
    args = command.split()
    return parser.parse_args(args=args)


def hash_file(file_object: IO):
    """
    Calculate MD5 hash of file.

    Args:
        file_object: File object

    Returns:
        MD5 hash of the file
    """
    # Calculate hash
    unique_hash = md5_hash(file_object.read())

    # Reset file pointer to start
    file_object.seek(0)

    return unique_hash


def num_files(path: Union[Path, str]) -> int:
    """
    Get the number of files in a path.

    Args:
        path: File path

    Returns:
        Number of files
    """
    return len(os.listdir(path))


def ngrams(tokens: List,
           n: int):
    """

    Args:
        tokens: List of elements
        n: N-gram size

    Returns:
        List of ngrams
    """
    return [tokens[i:i + n] for i in range(len(tokens) - n + 1)]


def print_json(d: Dict) -> None:
    """
    Render python dictionary as JSON with double quotes and indentation.

    Args:
        d: Python dictionary

    Returns:
        None
    """
    print(json.dumps(d, indent=4))


def read_as_base64(path: Union[str, Path]) -> str:
    """
    Convert file contents into a base64 string

    Args:
        path: File path

    Returns:
        Base64 string
    """
    content = Path(path).read_text()
    return base64.b64encode(content.encode('utf-8')).decode('utf-8')


def base64_dict(base64_str: str) -> Dict:
    """
    Parse a base64-encoded JSON as dictionary.

    Args:
        base64_str: Base-64 encoded string representation of JSON

    Returns:
        Python Dictionary
    """
    return json.loads(base64.b64decode(base64_str))


def format_as_hms(seconds: Union[int, float]) -> str:
    """
    Convert seconds to HH:MM:SS format.

    Args:
        seconds: Number of seconds

    Returns:
        String in the format HH:MM:SS
    """
    return time.strftime('%H:%M:%S', time.gmtime(seconds))


def roundup(n: float,
            m: int = 10) -> int:
    """
    Round up a number n to the nearest multiple of M.

    Args:
        n: Number
        m: Multiple of which number to roundup to

    Returns:
        Rounded integer number
    """
    return int(math.ceil(n / m)) * m


def minibatch(items, size):
    """
    Create mini-batches of length 'size' from a list of items.

    Original Source: `spacy` package

    Original function definition:
    https://github.com/explosion/spaCy/blob/master/spacy/util.py#L1426
    """
    if isinstance(size, int):
        size_ = itertools.repeat(size)
    else:
        size_ = size
    items = iter(items)
    while True:
        batch_size = next(size_)
        batch = list(itertools.islice(items, int(batch_size)))
        if len(batch) == 0:
            break
        yield list(batch)


def harmonic_mean(a: Union[int, float], b: Union[int, float]) -> Union[int, float]:
    """
    Compute harmonic mean of two numbers.

    Args:
        a: First number
        b: Second number

    Returns:
        Harmonic mean
    """
    return (2 * a * b) / (a + b)


def generate_edits(word: str, n: int = 1) -> List[str]:
    """
    Generate variations that are `n` edits away from word.

    Adapted from: https://norvig.com/spell-correct.html

    Args:
        word: Single word
        n: Number of edits away from word.

    Returns:
        List of edits
    """

    def edits1(word: str):
        letters = 'abcdefghijklmnopqrstuvwxyz'
        splits = [(word[:i], word[i:]) for i in range(len(word) + 1)]
        deletes = [L + R[1:] for L, R in splits if R]
        transposes = [L + R[1] + R[0] + R[2:] for L, R in splits if len(R) > 1]
        replaces = [L + c + R[1:] for L, R in splits if R for c in letters]
        inserts = [L + c + R for L, R in splits for c in letters]
        return set(deletes + transposes + replaces + inserts)

    edits = edits1(word)
    for i in range(n - 1):
        edits = [e2 for e1 in edits for e2 in edits1(e1)]
    return edits



import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin


def encode_histogram(text: str) -> np.ndarray:
    histogram = [0.0] * 256
    for ascii_code in text.encode('utf-8'):
        histogram[ascii_code] += 1
    return np.array(histogram) / len(text)


class HistogramEncoder(BaseEstimator, TransformerMixin):
    """
    Encode text into a bucket of 256 byte indices.
    """

    def __init__(self):
        super().__init__()

    def fit(self, *args, **kwargs):
        return self

    def transform(self, x):
        return np.stack([encode_histogram(e) for e in x])



def confusion_matrix_plot(y_true, y_pred) -> None:
    """
    Plot a confusion matrix.

    Args:
        y_true: List of true labels
        y_pred: List of prediction labels

    Returns:
    """
    from sklearn.metrics import ConfusionMatrixDisplay, confusion_matrix
    cm = confusion_matrix(y_true, y_pred)
    plot = ConfusionMatrixDisplay(confusion_matrix=cm).plot()
    plot.ax_.set_title('Confusion Matrix')



from typing import Tuple

import pandas as pd


def train_val_size(dataset,
                   val_ratio: float = 0.1) -> Tuple[int, int]:
    """
    Return the train and validation data sizes based on split ratio.

    Args:
        dataset: A python collection
        val_ratio: Ratio for validation dataset

    Returns:
        Tuple of number of rows for (training, validation)
    """
    val_size = int(val_ratio * len(dataset))
    train_size = len(dataset) - val_size
    return train_size, val_size


def view_result_table(cv) -> pd.DataFrame:
    """
    Display results from cross-validation.

    Args:
        cv: Result of cross-validation

    Returns:
        Pandas DataFrame
    """
    columns = ['params', 'mean_test_score', 'std_test_score', 'rank_test_score']
    return (pd.DataFrame(cv.cv_results_)[columns]
            .sort_values(by=['rank_test_score'])
            )


grid_report = view_result_table



from functools import lru_cache
from pathlib import Path
from typing import List

import pandas as pd
import numpy as np

NEGATION_WORDS = [
    "no",
    "not",
    "never",
    "none",
    "nothing",
    "nobody",
    "neither",
    "nowhere",
    "hardly",
    "scarcely",
    "barely",
    "doesn’t",
    "isn’t",
    "wasn’t",
    "shouldn’t",
    "wouldn’t",
    "couldn’t",
    "won’t",
    "can't",
    "don't",
]

INTERROGATIVE_WORDS = ["what", "who", "when", "where", "which", "why", "how"]

TENTATIVE_WORDS = [
    "appears to",
    "can",
    "could",
    "it is likely",
    "it is possible",
    "it is probable",
    "it is unlikely",
    "may",
    "might",
    "possibly",
    "probably",
    "seems to",
    "suggests that",
    "tends to",
]

REASON_WORDS = ["because", "reason", "as a result", "since", "therefore"]


@lru_cache(1)
def onegram_count() -> pd.DataFrame:
    """
    Get counts of 1-gram from Peter Norvig's list.

    Returns:
        DataFrame with one-gram, count and idf scores.
    """
    df = pd.read_csv(
        "https://norvig.com/ngrams/count_1w.txt",
        sep="\t",
        header=None,
        names=["word", "count"],
    )
    df["idf"] = np.log(df["count"].sum() / df["count"])
    df.sort_values(by="idf", ascending=True, inplace=True)
    return df


@lru_cache(1)
def bigram_count():
    # Fetch ngram counts from remote path
    url = "https://norvig.com/ngrams/count_2w.txt"
    bigram_df = pd.read_csv(url, sep="\t", names=["bigram", "count"])

    # Split bigram into separate columns
    bigram_df[["word1", "word2"]] = bigram_df.bigram.str.split(expand=True)
    return bigram_df


@lru_cache(1)
def nrc_vad() -> pd.DataFrame:
    url = "https://github.com/nchibana/moviearcs/raw/master/NRC-VAD-Lexicon.txt"
    return pd.read_csv(url, sep="\t")


def dict_words() -> List[str]:
    """
    Fetch default list of words present in Linux distros.

    Returns:
        List of words
    """
    return Path("/usr/share/dict/words").read_text().splitlines()



import json
import re
import string
from typing import Dict

from fns.json_encoders import NpEncoder

# Compiled regular expressions
_re_space = re.compile(r' {2,}')
_re_hashtag = re.compile(r'#')
_re_retweet = re.compile(r'^RT[\s]+')
_re_hyperlink = re.compile(r'https?:\/\/.*[\r\n]*')
_re_hyphen_word = re.compile('[a-zA-Z]+-[a-zA-Z]+')
_re_comma = re.compile(r',{2,}')


def combine_hyphenated_word(text: str) -> str:
    """
    Combine words in text that contain hyphen.

    Example: e-email to email

    Args:
        text: A sentence

    Returns:
        Processed sentence
    """
    return ' '.join(w.replace('-', '') if _re_hyphen_word.match(w) else w
                    for w in text.split())


def remove_hashtag(t: str) -> str:
    """
    Remove hashtag from the text.

    Args:
        t: Text

    Returns:
        Text without hashtag
    """
    return _re_hashtag.sub('', t)


def remove_retweet(t: str) -> str:
    """
    Remove RT from the text.

    Args:
        t: Text

    Returns:
        Text without RT symbol.
    """
    return _re_retweet.sub('', t)


def remove_hyperlink(t: str) -> str:
    """
    Remove hyperlinks from a text.

    Args:
        t: Text

    Returns:
        Text without hyperlinks.
    """
    return _re_hyperlink.sub('', t)


def remove_multiple_space(t: str) -> str:
    """
    Remove multiple spaces from the text.

    Adapted from: https://github.com/fastai/fastai/blob/master/fastai/text/core.py

    Args:
        t: Text

    Returns:
        Text without multiple space.
    """
    return _re_space.sub(' ', t)


def remove_multiple_commas(t: str) -> str:
    """
    Substitute multiple consecutive commas with a single comma.

    Usage:
    ```python
    >>> remove_multiple_commas('a,,b,c')
    'a,b,c'
    ```

    Args:
        t: Text

    Returns:
        Text without multiple commas.
    """
    return _re_comma.sub(',', t)


def remove_new_lines(text: str) -> str:
    """
    Strip away new lines at end.

    Args:
        t: Text

    Returns:
        Text without newline at end.
    """
    if isinstance(text, str):
        return text.replace('\\n', '').strip()
    return text


def remove_separator(text: str) -> str:
    """
    Keep only alphabet, number and space.

    Args:
        text:

    Returns:

    """
    no_separator_regex = re.compile(r'[^a-zA-Z0-9\s]')
    return no_separator_regex.sub('', text)


def remove_punctuation(text: str) -> str:
    """
    Remove all punctuations from a text.

    Args:
        text: Sentence

    """
    return ''.join(t for t in text if t not in string.punctuation)


def normalize_json(json_data: Dict) -> Dict:
    """
    Convert any non-standard types in dictionary to basic types.

    The normalization prevent errors during serialization.

    Usage:
    ```python
    >>> normalize_json({'nums': np.array([1, 2, 3]})
    {'nums': [1, 2, 3]}
    ```

    Args:
        json_data: Dictionary

    Returns:
        Normalized dictionary
    """
    return json.loads(json.dumps(json_data, cls=NpEncoder))



import json
from typing import Dict
from urllib.request import urlopen

IMAGENET_LABEL_TO_CLASS_URL = 'http://bit.ly/imagenet-labels'


def imagenet_index_to_class() -> Dict[int, str]:
    """
    Get a mapping from imagenet class index to class names.

    Returns:
        Mapping from imagenet class index to class names
    """
    raw_mapping = json.load(urlopen(IMAGENET_LABEL_TO_CLASS_URL))
    return {int(index): class_name for index, class_name in raw_mapping.items()}



import hashlib
import re
from typing import List, Tuple, Union

from sklearn.feature_extraction.text import TfidfVectorizer
import numpy as np
import pandas as pd


def md5_hash(text: str) -> str:
    """
    Generate MD5 hash of a text.

    Args:
        text: String

    Returns:
        MD5 hash
    """
    return hashlib.md5(text.encode('utf-8')).hexdigest()


def sha256hash(text: str) -> str:
    """
    Generate MD5 hash of a text.

    Args:
        text: String

    Returns:
        SHA256 hash
    """
    return hashlib.sha256(text.encode('utf-8')).hexdigest()


def window(tokens,
           size: int = 3):
    """
    Generate samples for a window size.

    Example:
    ```python
    >>> window(['a', 'b', 'c', 'd'], size=2)
    [(['a', 'b'], 'c'), (['b', 'c'], 'd')]
    ```

    Args:
        tokens: List of tokens
        size: Window size

    Returns:
        List of windowed samples
    """
    return [(tokens[i: i + size], tokens[i + size])
            for i in range(0, len(tokens) - size, 1)]


def offset_by_one(x,
                  sequence_length: int = 3):
    """
    Generate a list of small sequences offset by 1.

    Usage:

    ```python
    >>> offset_by_one([1, 2, 3, 4, 5], sequence_length=3)
    [([1, 2, 3], [2, 3, 4])]
    ```

    Args:
        x: Python list
        sequence_length: Chunk size

    Returns:

    """
    sl = sequence_length
    return [(x[i:i + sl], x[i + 1:i + sl + 1])
            for i in range(0, len(x) - sl - 1, sl)]


def num_words(text: str) -> int:
    """
    Counts the number of words using whitespace as delimiter.

    Args:
        text: Sentence

    Returns:
        Number of words
    """
    return len(text.split())


def unique_chars(texts: List[str]) -> List[str]:
    """
    Get a list of unique characters from list of text.

    Args:
        texts: List of sentences

    Returns:
        A sorted list of unique characters
    """
    return sorted(set(''.join(texts)))


def is_non_ascii(text: str) -> bool:
    """
    Check if text has non-ascci characters.

    Useful heuristic to find text containing emojis and non-english
    characters.

    Args:
        text: Sentence

    Returns:
        True if the text contains non-ascii characters.
    """
    try:
        text.encode('ascii')
        return False
    except UnicodeEncodeError:
        return True


def span_positions(text: str,
                   phrases: List[str]) -> List[Tuple[int, int]]:
    """
    Find span position of phrases in a text.

    Args:
        text: Sentence
        phrases: List of phrases

    Returns:
        List of span positions for each phrase.
        The span position is a tuple of start and end index.
    """
    capture_group = '|'.join([re.escape(phrase) for phrase in phrases])
    reg = re.compile(rf'\b({capture_group})\b', flags=re.IGNORECASE)
    return [match.span() for match in reg.finditer(text)]


def extract_abbreviations(texts: List[str]) -> List[str]:
    """
    Get a list of all-capitalized words.

    Example: WWW, HTTP, etc.

    Args:
        texts: List of sentences

    Returns:
        List of abbreviations
    """
    combined_text = '\n'.join(texts)
    symbols = re.findall(r'\b[A-Z][A-Z]+\b', combined_text)
    return list(set(symbols))


def export_fasttext_format(texts: List[str],
                           labels: Union[List[str], List[List[str]]],
                           filename) -> None:
    """
    Export training data to a fasttext compatible format.

    Format:
    __label__POSITIVE it was good

    Args:
        texts: List of sentences
        labels: List of single or multi-label classes
        filename: Exported filename

    Returns:
        None
    """
    output = []
    for text, text_label in zip(texts, labels):
        if type(text_label) is str:
            text_label = [text_label]
        labels = " ".join([f'__label__{label}' for label in text_label])
        output.append(f'{labels} {text}\n')
    with open(filename, 'w') as fp:
        fp.writelines(output)


def extract_tfidf_keywords(texts: List[str],
                           ngram: int = 2,
                           n: int = 10) -> List[str]:
    """
    Get top keywords based on mean tf-idf term score.

    Args:
        texts: List of sentences
        ngram: 1 for words, 2 for bigram and so on.
        n: Number of keywords to extract

    Returns:
        Keywords
    """
    tfidf = TfidfVectorizer(ngram_range=(1, ngram),
                            stop_words='english',
                            strip_accents='unicode',
                            sublinear_tf=True)
    vectors = tfidf.fit_transform(texts)
    term_tfidf = vectors.A.mean(axis=0)
    terms = np.array(tfidf.get_feature_names())
    return terms[term_tfidf.argsort()[::-1]][:n].tolist()


def extract_discriminative_keywords(df: pd.DataFrame,
                                    category_column: str,
                                    text_column: str,
                                    ngram: int = 2,
                                    n: int = 10) -> pd.DataFrame:
    """
    Generate discriminative keywords for texts in each category.

    Args:
        df: Dataframe with text and category columns.
        text_column: Column name containing texts
        category_column: Column name for the text category
        ngram: 1 for words, 2 for bigram and so on.
        n: Number of keywords to return.

    Returns:
        Dataframe with categories in columns and top-n keywords in each columns.
    """
    # Combine all texts into a single document for each category
    category_docs = df.groupby(by=category_column)[text_column].apply(' '.join)
    categories = category_docs.index.tolist()

    tfidf = TfidfVectorizer(ngram_range=(1, ngram),
                            stop_words='english',
                            strip_accents='unicode',
                            sublinear_tf=True)
    document_vectors = tfidf.fit_transform(category_docs).A
    keywords = np.array(tfidf.get_feature_names())
    top_terms = document_vectors.argsort(axis=1)[:, :n]
    return pd.DataFrame(keywords[top_terms].T,
                        columns=categories)


def extract_stopwords(texts: List[str]) -> pd.DataFrame:
    vec = TfidfVectorizer()
    vec.fit(texts)
    word_idf_pairs = zip(vec.get_feature_names(), vec.idf_)
    return (pd.DataFrame(word_idf_pairs, columns=["word", "idf"])
            .sort_values(by="idf"))



SENTENCE_BERT_STS_MODEL_NAME = 'distilbert-base-nli-stsb-mean-tokens'

CIFAR_10_CLASS_NAMES = ["airplane", "automobile", "bird", "cat", "deer",
                        "dog", "frog", "horse", "ship", "truck"]



import time
from functools import partial
from typing import List

from sklearn.dummy import DummyClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from sklearn.linear_model import LogisticRegression, SGDClassifier
from sklearn.model_selection import ParameterGrid, GridSearchCV, RandomizedSearchCV
from sklearn.multiclass import OneVsRestClassifier
from sklearn.naive_bayes import MultinomialNB, BernoulliNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import Pipeline
from sklearn.tree import DecisionTreeClassifier

from fns import format_as_hms
from fns.decorators import to
from fns.model_selection import view_result_table
import numpy as np
import os

__all__ = ['BaselineTextClassifier', 'generate_grid']


@to(list)
def generate_grid(grid,
                  multi_label: bool = False):
    for grid_class, param_grid in grid.items():
        for p in ParameterGrid(param_grid):
            if multi_label and (grid_class in [LogisticRegression, SGDClassifier, MultinomialNB, BernoulliNB]):
                yield OneVsRestClassifier(grid_class(**p))
            else:
                yield grid_class(**p)


grid = {
    'models': {LogisticRegression: {'class_weight': ['balanced'],
                                    'max_iter': [5000],
                                    'C': np.logspace(-4, 4, 50),
                                    'fit_intercept': [True, False],
                                    'solver': ['lbfgs']},
               SGDClassifier: {
                   'loss': ['hinge', 'log'],
                   'penalty': ['l1', 'l2', 'elasticnet'],
                   'alpha': np.logspace(-5, 1, 5),
                   'class_weight': ['balanced'],
                   'max_iter': [5000]
               },
               DecisionTreeClassifier: {
                   'class_weight': ['balanced']
               },
               RandomForestClassifier: {
                   'class_weight': ['balanced']
               }
               },
    'vectorizer': {TfidfVectorizer: {
        'ngram_range': [(1, 1), (1, 2), (1, 3), (2, 2), (2, 3)],
        'stop_words': [None, 'english'],
        'analyzer': ['word'],
        'binary': [True, False],
        'lowercase': [True, False],
        'max_df': [1.0, 0.75, 0.5, 0.25],
        'strip_accents': ['unicode', None],
    }}
}

baseline_grid = {
    'naive_bayes': {
        'models': {MultinomialNB: {}
                   },
        'vectorizer': {CountVectorizer: {
            'stop_words': [None, 'english'],
            'strip_accents': ['unicode', None],
        }}
    },
    'bernoulli_nb': {
        'models': {BernoulliNB: {}
                   },
        'vectorizer': {CountVectorizer: {
            'stop_words': [None, 'english'],
            'binary': [True],
            'strip_accents': ['unicode', None],
        }}
    },
    'dummy': {
        'models': {DummyClassifier: {
            'strategy': ['stratified', 'most_frequent', 'prior', 'uniform']
        }},
        'vectorizer': {CountVectorizer: {}
                       }
    },
    'knn': {
        'models': {KNeighborsClassifier: {
            'n_neighbors': [1, 3],
            'weights': ['uniform', 'distance'],
            'metric': ['euclidean', 'manhattan', 'minkowski']
        }
        },
        'vectorizer': {TfidfVectorizer: {}}
    },
}


def set_regularization_params(n_reg: int):
    raw_grid = grid.copy()
    raw_grid['models'][LogisticRegression]['C'] = np.logspace(-4, 4, n_reg)
    raw_grid['models'][SGDClassifier]['alpha'] = np.logspace(-5, 1, n_reg)
    return raw_grid


class BaselineTextClassifier:
    def __init__(self,
                 models: List = None,
                 vectorizers: List = None,
                 default_vectorizer: bool = True,
                 default_model: bool = True,
                 multi_label: bool = False,
                 n_reg: int = 50):
        self.multi_label = multi_label
        self.models = [] if not models else models
        self.vectorizers = [] if not vectorizers else vectorizers
        generate = partial(generate_grid, multi_label=multi_label)

        param_grid = set_regularization_params(n_reg=n_reg)

        if default_vectorizer:
            self.vectorizers += generate(param_grid['vectorizer'])
        if default_model:
            self.models += generate(param_grid['models'])

        self.params = []
        for model_name, model_grid in baseline_grid.items():
            self.params.append({'vectorizer': [*generate(model_grid['vectorizer'])],
                                'model': [*generate(model_grid['models'])],
                                })

        if self.vectorizers and self.models:
            self.params.append({'vectorizer': self.vectorizers,
                                'model': self.models,
                                })

    def fit(self,
            x_train,
            y_train,
            cv=3,
            scoring=None,
            memory=('/dev/shm/baseline' if os.name == 'posix' else None),
            search='grid',
            n_iter: int = 10,
            verbose: int = 1,
            n_jobs=-1):
        if scoring is None:
            scoring = 'f1_samples' if self.multi_label else 'f1_macro'
        classifier_pipeline = Pipeline([('vectorizer', TfidfVectorizer()),
                                        ('model', OneVsRestClassifier(LogisticRegression()))],
                                       memory=memory)

        common_params = dict(estimator=classifier_pipeline,
                             cv=cv,
                             verbose=verbose,
                             refit=True,
                             scoring=scoring,
                             n_jobs=n_jobs)

        if search == 'grid':
            hpo = GridSearchCV(param_grid=self.params,
                               **common_params)
        else:
            hpo = RandomizedSearchCV(param_distributions=self.params,
                                     n_iter=n_iter,
                                     **common_params)

        hpo.fit(x_train, y_train)
        print(f'Best F1-score: {hpo.best_score_}')
        print(f'Best Estimator: {hpo.best_estimator_}')
        scores_df = view_result_table(hpo)
        return scores_df

    def timeit(self, x_train, y_train, n_iter: int = 10) -> None:
        """
        Estimate time for a full-grid search.

        Args:
            x_train: Training texts
            y_train: Training labels
            n_iter: Number of iterations used to estimate time

        Returns:

        """
        total_tasks = sum(len(ParameterGrid(p)) for p in self.params) * 3
        start_time = time.perf_counter()
        self.fit(x_train, y_train, search='random', n_iter=n_iter)
        time_taken = (time.perf_counter() - start_time) / n_iter
        total_time = time_taken * total_tasks
        print(f'Estimated time for full grid: {format_as_hms(total_time)}')



from functools import partial
from typing import List, Optional

import optuna
from sklearn.dummy import DummyClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.linear_model import LogisticRegression, SGDClassifier
from sklearn.model_selection import ParameterGrid, cross_val_score
from sklearn.multiclass import OneVsRestClassifier
from sklearn.naive_bayes import MultinomialNB, BernoulliNB
from sklearn.neural_network import MLPClassifier
from sklearn.pipeline import Pipeline
from sklearn.tree import DecisionTreeClassifier

from fns.decorators import to

__all__ = ['BaselineTextClassifier', 'generate_grid']


@to(list)
def generate_grid(grid,
                  multi_label: bool = False):
    for grid_class, param_grid in grid.items():
        for p in ParameterGrid(param_grid):
            if multi_label and (grid_class in [LogisticRegression, SGDClassifier, MultinomialNB, BernoulliNB]):
                yield OneVsRestClassifier(grid_class(**p))
            else:
                yield grid_class(**p)


def evaluate_trial(trial,
                   x,
                   y,
                   cv,
                   scoring,
                   vectorizers,
                   multi_label=False):
    model_class = trial.suggest_categorical('model_class', ['LogisticRegression',
                                                            'SGDClassifier',
                                                            'DecisionTreeClassifier',
                                                            'RandomForestClassifier',
                                                            'DummyClassifier',
                                                            'MLPClassifier'])
    if vectorizers is None:
        vectorizers = [TfidfVectorizer()]
    else:
        vectorizers += [TfidfVectorizer()]
    vectorizer_choices = {v.__class__.__name__: v for v in vectorizers}
    vectorizer_name = trial.suggest_categorical('vectorizer_name',
                                                list(vectorizer_choices))

    if vectorizer_name == 'TfidfVectorizer':
        ngram_range = trial.suggest_int('ngram_range', 1, 3)
        vectorizer_params = {
            'ngram_range': (1, ngram_range),
            'stop_words': trial.suggest_categorical('stop_words', [None, 'english']),
            'analyzer': 'word',
            'binary': trial.suggest_categorical('binary', [True, False]),
            'lowercase': trial.suggest_categorical('lowercase', [True, False]),
            'sublinear_tf': trial.suggest_categorical('sublinear_tf', [True, False]),
            'max_df': trial.suggest_discrete_uniform('max_df', 0.2, 1.0, 0.1),
            'strip_accents': trial.suggest_categorical('strip_accents', ['unicode', None]),
        }
        vectorizer = TfidfVectorizer(**vectorizer_params)
    else:
        vectorizer = vectorizer_choices[vectorizer_name]

    if model_class == 'LogisticRegression':
        base_class = LogisticRegression
        params = {
            'C': trial.suggest_loguniform('C', 1e-5, 10),
            'class_weight': 'balanced',
            'max_iter': 5000,
            'fit_intercept': trial.suggest_categorical('fit_intercept', [True, False])
        }
    elif model_class == 'SGDClassifier':
        base_class = SGDClassifier
        params = {
            'loss': trial.suggest_categorical('loss', ['hinge', 'log']),
            'penalty': trial.suggest_categorical('penalty', ['l1', 'l2', 'elasticnet']),
            'alpha': trial.suggest_loguniform('alpha', 1e-5, 10),
            'class_weight': 'balanced',
            'max_iter': 5000
        }
    elif model_class == 'DecisionTreeClassifier':
        base_class = DecisionTreeClassifier
        params = {'class_weight': 'balanced'}
    elif model_class == 'RandomForestClassifier':
        base_class = RandomForestClassifier
        params = {'class_weight': 'balanced'}
    elif model_class == 'MLPClassifier':
        base_class = MLPClassifier
        params = {
            'activation': trial.suggest_categorical('activation', ['relu', 'tanh', 'identity', 'logistic']),
            'solver': 'adam',
            'alpha': trial.suggest_loguniform('alpha', 1e-4, 10),
            'early_stopping': trial.suggest_categorical('early_stopping', [True, False]),
        }
    else:
        base_class = DummyClassifier
        params = {'strategy': trial.suggest_categorical('strategy',
                                                        ['stratified', 'most_frequent', 'prior', 'uniform'])
                  }

    model = base_class(**params)
    if multi_label and (base_class in [LogisticRegression, SGDClassifier]):
        model = OneVsRestClassifier(model)

    pipeline = Pipeline([('vectorizer', vectorizer),
                         ('model', model)])

    return cross_val_score(pipeline,
                           x,
                           y,
                           cv=cv,
                           scoring=scoring).mean()


class BaselineTextClassifier:
    def __init__(self,
                 study_name: str = 'baseline',
                 reset=False):
        if reset:
            import os
            os.remove(f'{study_name}.db')
        self.study = optuna.create_study(study_name=study_name,
                                         storage=f'sqlite:///{study_name}.db',
                                         direction='maximize',
                                         load_if_exists=True)

    def fit(self,
            x: List[str],
            y,
            cv: int = 5,
            scoring: str = 'f1_macro',
            multi_label: bool = False,
            vectorizers: Optional[List] = None,
            n_trials: Optional[int] = None,
            timeout: Optional[int] = None):
        optimize_function = partial(evaluate_trial,
                                    x=x,
                                    y=y,
                                    cv=cv,
                                    scoring=scoring,
                                    vectorizers=vectorizers,
                                    multi_label=multi_label)
        self.study.optimize(optimize_function,
                            n_trials=n_trials,
                            timeout=timeout)



import subprocess
import os
import time
from importlib import import_module


def run_foreground(cmd: str) -> None:
    """
    Run a bash command in foreground.

    Reference: http://blog.kagesenshi.org/2008/02/teeing-python-subprocesspopen-output.html

    Args:
        cmd: Bash command

    Returns:
        None
    """
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    while True:
        line = p.stdout.readline()
        print(line.strip())
        if line == '' and p.poll() != None:
            break
    return None


def run_background(command: str) -> None:
    """
    Run a bash command in background.

    Args:
        command: Bash command

    Returns:
        None
    """
    subprocess.Popen(command, shell=True)


def jupyter() -> None:
    """
    Start a jupyter lab server using localtunnel.

    Returns:
        None
    """
    from google.colab import drive
    drive.mount('/content/drive')
    os.chdir('/content/drive/MyDrive/colab/')
    run_background('nohup pip install jupyterlab --upgrade -qqq &')
    command = 'nohup jupyter lab --notebook-dir="/content/drive/MyDrive/colab" --no-browser --allow-root --ip="0.0.0.0" --port="6006" &'
    run_background(command)
    time.sleep(5)
    run_foreground('npx localtunnel --port 6006')


def vscode(subdomain: str = 'amitness',
           port: int = 9000,
           config_save_path: str = '/content/drive/MyDrive/colab/.vscode') -> None:
    """
    Start VSCode server which persists all settings and extensions.

    Args:
        subdomain: Subdomain for localtunnel.
        port: Port for running code-server
        config_save_path: Path in Google Drive to save VSCode settings

    Returns:
        None
    """
    drive = import_module('google.colab.drive')
    drive.mount('/content/drive')
    subprocess.run(['curl', '-fsSL', 'https://code-server.dev/install.sh', '-O'])
    subprocess.run(['bash', 'install.sh', '--version', '3.10.2'])
    subprocess.run(['pip3', 'install', 'flake8', '--user'])
    subprocess.run(['pip3', 'install', 'black', '--user'])
    print(f'https://{subdomain}.loca.lt/?folder=/content/drive/MyDrive/colab')
    run_foreground(
        f'code-server --port {port} --auth none --disable-telemetry --force --user-data-dir {config_save_path} & npx localtunnel -p {port} -s {subdomain} --allow-invalid-cert')


def expose_port(port: int, path: str = '/') -> None:
    """
    Expose port as an external URL.

    The URL is only accessible to you and available till the notebook runs.

    Args:
        port: Port a service is running on
        path: Path the service is running on

    Returns:
        None
    """
    output = import_module('google.colab.output')
    output.serve_kernel_port_as_window(port, path=path)



from setuptools import setup, find_packages

setup(
    name='fns',
    version='0.5.19',
    license='MIT',
    description='Reusable functions for use in ML projects',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.7",
    author='Amit Chaudhary',
    author_email='meamitkc@gmail.com',
    url='https://github.com/amitness/fns',
    install_requires=['numpy'],
    packages=find_packages(),
)



from typing import List, Union

import pandas as pd
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.pipeline import make_pipeline, FeatureUnion
from sklearn.preprocessing import FunctionTransformer

from fns.metrics import n_clusters


def cluster_text(texts: List[str],
                 return_dataframe: bool = True,
                 n: int = None) -> Union[pd.DataFrame, List[int]]:
    """
    Quickly cluster a list of sentences for EDA.

    Args:
        n: Number of clusters
        texts: List of sentences
        return_dataframe: Whether to return as dataframe or a list of cluster labels

    Returns:

    """
    n = n or n_clusters(texts)
    hybrid_tfidf = FeatureUnion([('word_tfidf', TfidfVectorizer(ngram_range=(1, 2),
                                                                strip_accents='unicode',
                                                                analyzer='word',
                                                                stop_words='english',
                                                                sublinear_tf=True)),
                                 ('char_tfidf', TfidfVectorizer(ngram_range=(3, 3),
                                                                strip_accents='unicode',
                                                                analyzer='char_wb',
                                                                stop_words='english',
                                                                sublinear_tf=True))])
    cluster_pipeline = make_pipeline(hybrid_tfidf,
                                     FunctionTransformer(lambda x: x.todense(), accept_sparse=True, validate=False),
                                     PCA(0.9, random_state=0),
                                     KMeans(n, random_state=0))
    clusters = cluster_pipeline.fit_predict(texts)
    if return_dataframe:
        return (pd.DataFrame({'text': texts, 'cluster': clusters})
                .assign(cluster_size=lambda d: d['cluster'].map(d['cluster'].value_counts()))
                .sort_values(by=['cluster_size', 'cluster', 'text'], ascending=[False, True, True])
                .drop(columns=['cluster_size']))
    else:
        return clusters


def similarity_sort(texts: List[str]) -> List[str]:
    """
    Sort list of sentences such that similar sentence are placed consecutively.

    Args:
        texts: List of sentences

    Returns:
        Sorted sentences
    """
    df = cluster_text(texts, n=len(texts) // 2)
    return df['text'].tolist()
